structure PosixFileSys: POSIX_FILESYS_EXTRA =
   struct
      (* Patch to make Time look like it deals with Int.int
       * instead of LargeInt.int.
       *)
      structure Time =
	 struct
	    open Time
	    fun toSeconds a = (LargeInt.toInt o Time.toSeconds) a
	    fun fromSeconds a = (Time.fromSeconds o LargeInt.fromInt) a
	 end
      
      structure Error = PosixError
      structure Prim = PosixPrimitive.FileSys
      open Prim
      structure Stat = Prim.Stat
      structure Flags = PosixFlags
      structure CString = PosixPrimitive.CString
      structure ForeignPtr = PosixPrimitive.ForeignPtr

      val R_OK = R_OK_
      val W_OK = W_OK_
      val X_OK = X_OK_
      val F_OK = F_OK_

      val checkResult = Error.checkResult

      type file_desc = Prim.file_desc
      type uid = Prim.uid
      type gid = Prim.gid

      fun fdToWord n = SysWord.fromInt n
      val wordToFD = SysWord.toInt
      fun id x = x
      val fdToIOD = id
      fun iodToFD a = SOME a

      (*------------------------------------*)
      (*             dirstream              *)
      (*------------------------------------*)

      local
	 structure Prim = Prim.Dirstream
	 datatype dirstream = DS of Prim.dirstream option ref

	 fun get (DS r) =
	    case !r of
	       NONE => Error.raiseSys Error.badf
	     | SOME d => d
      in
	 type dirstream = dirstream
	    
	 fun opendir s =
	    let val d = Prim.opendir s
	    in if ForeignPtr.isNull d then Error.error ()
	       else DS (ref (SOME d))
	    end

	 fun readdir d =
	    let
	       val d = get d
	       fun loop () =
		  let
		     val _ = Error.clearErrno ()
		     val cs = Prim.readdir d
		  in if CString.isNull cs
			then if Error.getErrno () = 0
				then ""
			     else Error.error ()
		     else (case CString.toString cs of
			      "." => loop ()
			    | ".." => loop ()
			    | s => s)
		  end
	    in loop ()
	    end

	 fun rewinddir d =
	    let val d = get d
	    in Error.clearErrno ()
	       ; Prim.rewinddir d
	       ; if Error.getErrno () = 0 then () else Error.error ()
	    end

	 fun closedir (DS r) =
	    case !r of
	       NONE => ()
	     | SOME d => (checkResult (Prim.closedir d); r := NONE)
      end
	 
      fun chdir s =
	 checkResult (Prim.chdir s)

      fun getcwd () = Prim.getcwd ()
	 
      val stdin = 0
      val stdout = 1
      val stderr = 2

      structure S =
	 struct
	    open S Flags
	 end

      structure O =
	 struct
	    open O Flags
	 end

      datatype open_mode = O_RDONLY | O_WRONLY | O_RDWR

      fun wordToOpenMode w =
	 if w = o_rdonly then O_RDONLY
	 else if w = o_wronly then O_WRONLY
	      else if w = o_rdwr then O_RDWR
		   else raise Fail "wordToOpenMode: unknown word"
		      
      val openModeToWord =
	 fn O_RDONLY => o_rdonly
	  | O_WRONLY => o_wronly
	  | O_RDWR => o_rdwr

      val error = PosixError.error

      fun createf (pathname, openMode, flags, mode) =
	 let
	    val fd =
	       Prim.openn (pathname,
			   Flags.flags [openModeToWord openMode, flags, O.creat],
			   mode)
	 in if fd = ~1
	       then error ()
	    else fd
	 end

      fun openf (pathname, openMode, flags) =
	 let val fd = Prim.openn (pathname,
				  Flags.flags [openModeToWord openMode, flags],
				  Flags.empty)
	 in if fd = ~1
	       then error ()
	    else fd
	 end
	 
      fun creat (s, m) = createf (s, O_WRONLY, O.trunc, m)

      val umask = Prim.umask

      local
	 fun wrap p arg = (checkResult (p arg); ())
      in
	 fun link a      = wrap Prim.link a
	 fun mkdir a     = wrap Prim.mkdir a
	 fun mkfifo a    = wrap Prim.mkfifo a
	 fun unlink a    = wrap Prim.unlink a
	 fun rmdir a     = wrap Prim.rmdir a
	 fun rename a    = wrap Prim.rename a
	 fun symlink a   = wrap Prim.symlink a
	 fun chmod a     = wrap Prim.chmod a
	 fun fchmod a    = wrap Prim.fchmod a
	 fun chown a     = wrap Prim.chown a
	 fun fchown a    = wrap Prim.fchown a
	 fun ftruncate a = wrap Prim.ftruncate a
      end	    

      fun readlink (path: string): string =
	let val cs = Prim.readlink path
	in if CString.isNull cs then Error.error()
	   else CString.toString cs
	end

      type dev = Prim.dev
      val wordToDev = id
      val devToWord = id

      type ino = Prim.ino
      val wordToIno = SysWord.toInt
      val inoToWord = SysWord.fromInt

      structure ST =
	 struct
	    datatype stat =
	       T of {dev: dev,
		     ino: ino,
		     mode: S.mode,
		     nlink: int,
		     uid: uid,
		     gid: gid,
		     size: int,
		     atime: Time.time,
		     mtime: Time.time,
		     ctime: Time.time}

	    fun fromC (): stat =
	       T {dev = Stat.dev (),
		  ino = Stat.ino (),
		  mode = Stat.mode (),
		  nlink = Stat.nlink (),
		  uid = Stat.uid (),
		  gid = Stat.gid (),
		  size = Stat.size (),
		  atime = Time.fromSeconds (Stat.atime ()),
		  mtime = Time.fromSeconds (Stat.mtime ()),
		  ctime = Time.fromSeconds (Stat.ctime ())}

	    local
	       fun make sel (T r) = sel r
	    in
	       fun mode a  = make #mode a
	       fun ino a   = make #ino a
	       fun dev a   = make #dev a
	       fun nlink a = make #nlink a
	       fun uid a   = make #uid a
	       fun gid a   = make #gid a
	       fun size a  = make #size a
	       fun atime a = make #atime a
	       fun mtime a = make #mtime a
	       fun ctime a = make #ctime a
	    end

	    local
	       fun make prim s = prim (mode s)
	    in
	       fun isDir a  = make Prim.ST.isDir a
	       fun isChr a  = make Prim.ST.isChr a
	       fun isBlk a  = make Prim.ST.isBlk a
	       fun isReg a  = make Prim.ST.isReg a
	       fun isFIFO a = make Prim.ST.isFIFO a
	       fun isLink a = make Prim.ST.isLink a
	       fun isSock a = make Prim.ST.isSock a
	    end
	 end

      local
	 fun make prim arg =
	    (checkResult (prim arg)
	     ; ST.fromC ())
      in
	 fun stat a  = make Prim.Stat.stat a
	 fun lstat a = make Prim.Stat.lstat a
	 fun fstat a = make Prim.Stat.fstat a
      end

      datatype access_mode = A_READ | A_WRITE | A_EXEC

      val conv_access_mode =
	 fn A_READ => R_OK
	  | A_WRITE => W_OK
	  | A_EXEC => X_OK

      fun access (path: string, mode: access_mode list): bool =
	 let val mode = Flags.flags (F_OK :: (map conv_access_mode mode))
	 in case Prim.access (path, mode) of
	    ~1 => false
	  | _ => true
	 end

      local
	 structure U = Prim.Utimbuf
      in
	 fun utime (f: string, opt: {actime: Time.time,
				    modtime: Time.time} option): unit =
	    let
	       val (a, m) =
		  case opt of
		     NONE => let val t = Time.now ()
			     in (t, t)
			     end
		   | SOME {actime = a, modtime = m} => (a, m)
	       val a = Time.toSeconds a
	       val m = Time.toSeconds m
	    in U.setActime a
	       ; U.setModtime m
	       ; checkResult (U.utime f)
	    end
      end

      local
	 fun convertProperty s =
	   case propertyLookup s of
	     NONE => Error.raiseSys Error.inval
	   | SOME n => n

	 fun make prim (f, s) =
	    let val n = prim (f, convertProperty s)
	    in if n < 0
		  then Error.error ()
	       else if n = 0
		       then NONE
		    else SOME (SysWord.fromInt n)
	    end
      in
	 fun pathconf a  = make Prim.pathconf a
	 fun fpathconf a = make Prim.fpathconf a
      end
   end
