structure PosixIO: POSIX_IO =
   struct
      structure Prim = PosixPrimitive.IO
      open Prim
      structure Error = PosixError
      val checkResult = Error.checkResult
      val checkReturnResult = Error.checkReturnResult
      structure FS = PosixFileSys
      structure CString = PosixPrimitive.CString

      type file_desc = Prim.file_desc
      type pid = Prim.pid

      val F_DUPFD = Prim.F_DUPFD_
      val F_GETFD = Prim.F_GETFD_
      val F_SETFD = Prim.F_SETFD_
      val F_GETFL = Prim.F_GETFL_
      val F_SETFL = Prim.F_SETFL_
      val F_GETLK = Prim.F_GETLK_
      val F_SETLK = Prim.F_SETLK_
      val F_RDLCK = Prim.F_RDLCK_
      val F_WRLCK = Prim.F_WRLCK_
      val F_UNLCK = Prim.F_UNLCK_
      val F_SETLKW = Prim.F_SETLKW_
      val F_GETOWN = Prim.F_GETOWN_
      val F_SETOWN = Prim.F_SETOWN_
      val O_ACCMODE = Prim.O_ACCMODE_
      val SEEK_SET = Prim.SEEK_SET_
      val SEEK_CUR = Prim.SEEK_CUR_
      val SEEK_END = Prim.SEEK_END_

      val pipe = fn x => case pipe x
			   of SOME r => r
			    | NONE => Error.error()

      fun dup fd = checkReturnResult (Prim.dup fd)

      fun dup2 {old, new} = checkResult (Prim.dup2 (old, new))

      fun close fd = checkResult (Prim.close fd)

      fun checkSlice (buf_len:int, i:int, sz:int option) : int =
	case sz
	  of NONE => if i < 0 orelse i > buf_len then raise Subscript
		     else buf_len - i
	   | SOME n => if i < 0 orelse n < 0 orelse i > buf_len - n then raise Subscript
		       else n
	
      fun readArr (fd, {buf:Word8Array.array, i:int, sz:int option}): int =
	let val num = checkSlice(Word8Array.length buf, i, sz)
	in checkReturnResult (Prim.readArr (fd, buf, i, num))
	end

      fun readVec (fd, n): Word8Vector.vector =
	if n < 0 then raise Subscript
	else case Prim.readVec(fd, n)
	       of SOME v => v
		| NONE => Error.error()
	 
      fun writeVec (fd, {buf:Word8Vector.vector, i:int, sz:int option}) : int =
	let val num = checkSlice (Word8Vector.length buf, i, sz)
	in checkReturnResult (Prim.writeVec (fd, buf, i, num))
	end

      fun writeArr (fd, {buf:Word8Array.array, i:int, sz:int option}) : int =
	let val num = checkSlice (Word8Array.length buf, i, sz)
	in checkReturnResult (Prim.writeArr (fd, buf, i, num))
	end

      structure FD =
	struct
	  open FD PosixFlags
	end

      structure O = PosixFileSys.O

      datatype open_mode = datatype PosixFileSys.open_mode
	 
      fun dupfd {old, base} =
	checkReturnResult (Prim.fcntl3 (old, F_DUPFD, base))

      fun getfd fd =
	Word.fromInt (checkReturnResult (Prim.fcntl2 (fd, F_GETFD)))

      fun setfd (fd, flags): unit =
	checkResult (Prim.fcntl3 (fd, F_SETFD, Word.toInt flags))
			    
      fun getfl fd : O.flags * open_mode =
	 let val n = Prim.fcntl2 (fd, F_GETFD)
	 in if n < 0
	       then Error.error ()
	    else let val w = Word.fromInt n
		     val flags = Word.andb (w, Word.notb O_ACCMODE)
		     val mode = Word.andb (w, O_ACCMODE)
		 in (flags, PosixFileSys.wordToOpenMode mode)
		 end
	 end
      
      fun setfl (fd, flags: O.flags): unit  =
	checkResult (Prim.fcntl3 (fd, F_SETFL, Word.toInt flags))
	 
      datatype whence = SEEK_SET | SEEK_CUR | SEEK_END

      val whenceToInt =
	fn SEEK_SET => Prim.SEEK_SET_
	 | SEEK_CUR => Prim.SEEK_CUR_
	 | SEEK_END => Prim.SEEK_END_

      fun intToWhence n =
	if n = Prim.SEEK_SET_ then SEEK_SET
	else if n = Prim.SEEK_CUR_ then SEEK_CUR
	     else if n = Prim.SEEK_END_ then SEEK_END
		  else raise Fail "Posix.IO.intToWhence"
		      
      fun lseek (fd, n: Position.int, w: whence): Position.int =
	checkReturnResult (Prim.lseek (fd, n, whenceToInt w))
	 
      fun fsync fd : unit = checkResult (Prim.fsync fd)
	 
      datatype lock_type = F_RDLCK | F_WRLCK | F_UNLCK

      val lockTypeToInt =
	fn F_RDLCK => Prim.F_RDLCK_
	 | F_WRLCK => Prim.F_WRLCK_
	 | F_UNLCK => Prim.F_UNLCK_

      fun intToLockType n =
	if n = Prim.F_RDLCK_ then F_RDLCK
	else if n = Prim.F_WRLCK_ then F_WRLCK
	     else if n = Prim.F_UNLCK_ then F_UNLCK
		  else raise Fail "Posix.IO.intToLockType"
	 
      structure FLock =
	struct
	  type flock = {ltype: lock_type,
			whence: whence,
			start: Position.int,
			len: Position.int,
			pid: pid option}
			 
	  fun flock l = l
	  fun ltype (f: flock) : lock_type = #ltype f
	  fun whence (f: flock) : whence = #whence f
	  fun start (f: flock) : Position.int = #start f
	  fun len (f: flock) : Position.int = #len f
	  fun pid (f: flock) : pid option = #pid f
	end

      local
	 structure P = Prim.FLock
	 fun make 
	   (cmd, usepid)
	   (fd, {ltype, whence, start, len, ...}: FLock.flock)
	   : FLock.flock  =
	    (P.setType (lockTypeToInt ltype)
	     ; P.setWhence (whenceToInt whence)
	     ; P.setStart start
	     ; P.setLen len
	     ; checkResult (P.fcntl (fd, cmd))
	     ; {ltype = intToLockType (P.typ ()),
		whence = intToWhence (P.whence ()),
		start = P.start (),
		len = P.len (),
		pid = if usepid then SOME (P.pid ()) else NONE})
      in
	fun getlk a  = make (F_GETLK, true) a
	fun setlk a  = make (F_SETLK, false) a
	fun setlkw a = make (F_SETLKW, false) a
      end
   end
